@@ -45,13 +45,7 @@
 #include <algorithm>
 #include <fcntl.h>
 #include <thread>
-
-#ifdef INCLUDE_CLIENT_APIS
-SOCKET gSsdpReqSocket4 = INVALID_SOCKET;
-#ifdef UPNP_ENABLE_IPV6
-SOCKET gSsdpReqSocket6 = INVALID_SOCKET;
-#endif /* UPNP_ENABLE_IPV6 */
-#endif /* INCLUDE_CLIENT_APIS */
+#include <vector>
 
 // Extract criteria from ssdp packet. Cmd can come from either an USN,
 // NT, or ST field. The possible forms are:
@@ -340,7 +334,7 @@ static int create_ssdp_sock_v4(SOCKET *ssdpSock)
         goto error_handler;
     }
 
-    for (const auto& netif : g_netifs) {
+    for (const auto& netif: g_netifs) {
         auto ipaddr = netif.firstipv4addr();
         if (!ipaddr)
             continue;
@@ -364,10 +358,10 @@ static int create_ssdp_sock_v4(SOCKET *ssdpSock)
     return UPNP_E_SUCCESS;
 
 error_handler:
-    char errorBuffer[ERROR_BUFFER_LEN];
-    posix_strerror_r(errno, errorBuffer, ERROR_BUFFER_LEN);
+    std::string errorDesc;
+    NetIF::getLastError(errorDesc);
     UpnpPrintf(UPNP_CRITICAL, SSDP, __FILE__, __LINE__,
-               "%s: %s\n", errorcause.c_str(), errorBuffer);
+               "%s: %s\n", errorcause.c_str(), errorDesc.c_str());
     if (*ssdpSock != INVALID_SOCKET) {
         UpnpCloseSocket(*ssdpSock);
     }
@@ -387,20 +381,15 @@ static int sock_make_no_blocking(SOCKET sock)
 #endif /* !_WIN32 */
 }
 
-/* Create the SSDP IPv4 socket to be used by the control point. */
-static int create_ssdp_sock_reqv4(SOCKET *ssdpReqSock, int port)
+// Create an IPV4 multicast datagram socket for the CP to emit search requests and receive
+// unicast responses on a given interface (designated by sadrv4)
+static int create_ssdp_sock_reqv4(const std::string& sadrv4, SOCKET* ssdpReqSock, int port)
 {
     char ttl = 2;
     int ret = UPNP_E_SOCKET_ERROR;
 
     *ssdpReqSock = INVALID_SOCKET;
 
-    std::string sadrv4 = apiFirstIPV4Str();
-    if (sadrv4.empty()) {
-        UpnpPrintf(UPNP_ERROR, SSDP, __FILE__, __LINE__, "create_ssdp_sock_reqv4: no IPV4??\n");
-        return ret;
-    }
-
     uint32_t hostaddrv4;
     std::string errorcause;
 
@@ -428,8 +417,7 @@ static int create_ssdp_sock_reqv4(SOCKET *ssdpReqSock, int port)
 
     sock_make_no_blocking(*ssdpReqSock);
 
-    if (port > 0)
-    {
+    if (port > 0) {
         struct sockaddr_storage ss = {};
         auto ssdpAddr4 = reinterpret_cast<struct sockaddr_in *>(&ss);
 
@@ -448,9 +436,10 @@ static int create_ssdp_sock_reqv4(SOCKET *ssdpReqSock, int port)
     return UPNP_E_SUCCESS;
 
 error_handler:
-    char errorBuffer[ERROR_BUFFER_LEN];
-    posix_strerror_r(errno, errorBuffer, ERROR_BUFFER_LEN);
-    UpnpPrintf(UPNP_CRITICAL, SSDP, __FILE__, __LINE__, "%s: %s\n", errorcause.c_str(), errorBuffer);
+    std::string errorDesc;
+    NetIF::getLastError(errorDesc);
+    UpnpPrintf(UPNP_CRITICAL, SSDP, __FILE__, __LINE__, "%s: %s\n", errorcause.c_str(),
+               errorDesc.c_str());
     if (*ssdpReqSock != INVALID_SOCKET) {
         UpnpCloseSocket(*ssdpReqSock);
     }
@@ -529,9 +518,10 @@ static int create_ssdp_sock_v6(bool isulagua, SOCKET *ssdpSock)
     return UPNP_E_SUCCESS;
 
 error_handler:
-    char errorBuffer[ERROR_BUFFER_LEN];
-    posix_strerror_r(errno, errorBuffer, ERROR_BUFFER_LEN);
-    UpnpPrintf(UPNP_CRITICAL, SSDP, __FILE__, __LINE__, "%s: %s\n", errorcause.c_str(), errorBuffer);
+    std::string errorDesc;
+    NetIF::getLastError(errorDesc);
+    UpnpPrintf(UPNP_CRITICAL, SSDP, __FILE__, __LINE__, "%s: %s\n", errorcause.c_str(),
+               errorDesc.c_str());
     if (*ssdpSock != INVALID_SOCKET) {
         UpnpCloseSocket(*ssdpSock);
     }
@@ -540,15 +530,15 @@ error_handler:
 
 
 #ifdef INCLUDE_CLIENT_APIS
-/* Create the SSDP IPv6 socket to be used by the control point. */
-static int create_ssdp_sock_reqv6(SOCKET *ssdpReqSock, int port)
+// Create an IPV6 multicast datagram socket for the CP to emit search requests and receive unicast
+// responses on a given interface, designated by its index.
+static int create_ssdp_sock_reqv6(int index, SOCKET* ssdpReqSock, int port)
 {
 #ifdef _WIN32
     DWORD hops = 1;
 #else
     int hops = 1;
 #endif
-    int index = apiFirstIPV6Index();
     std::string errorcause;
     int ret = UPNP_E_SOCKET_ERROR;
 
@@ -573,13 +563,12 @@ static int create_ssdp_sock_reqv6(SOCKET *ssdpReqSock, int port)
 
     sock_make_no_blocking(*ssdpReqSock);
 
-    if (port > 0)
-    {
+    if (port > 0) {
         int onOff = 1;
 
         // Set IPV6 socket to only bind to IPV6 (Linux Dual Stack)
         ret = setsockopt(*ssdpReqSock, IPPROTO_IPV6, IPV6_V6ONLY,
-                reinterpret_cast<char *>(&onOff), sizeof(onOff));
+                         reinterpret_cast<char *>(&onOff), sizeof(onOff));
 
         if (ret == -1) {
             errorcause = "setsockopt() IPV6_V6ONLY";
@@ -605,9 +594,10 @@ static int create_ssdp_sock_reqv6(SOCKET *ssdpReqSock, int port)
     return UPNP_E_SUCCESS;
 
 error_handler:
-    char errorBuffer[ERROR_BUFFER_LEN];
-    posix_strerror_r(errno, errorBuffer, ERROR_BUFFER_LEN);
-    UpnpPrintf(UPNP_CRITICAL, SSDP, __FILE__, __LINE__, "%s: %s\n", errorcause.c_str(), errorBuffer);
+    std::string errorDesc;
+    NetIF::getLastError(errorDesc);
+    UpnpPrintf(UPNP_CRITICAL, SSDP, __FILE__, __LINE__, "%s: %s\n", errorcause.c_str(),
+               errorDesc.c_str());
     if (*ssdpReqSock != INVALID_SOCKET) {
         UpnpCloseSocket(*ssdpReqSock);
     }
@@ -618,52 +608,109 @@ error_handler:
 #endif /* UPNP_ENABLE_IPV6 */
 
 
-static void maybeCLoseAndInvalidate(SOCKET *s, int doclose)
+static void maybeCLoseAndInvalidate(SOCKET& s, int doclose)
 {
-    if (doclose && *s != INVALID_SOCKET) {
-        UpnpCloseSocket(*s);
+    if (doclose && s != INVALID_SOCKET) {
+        UpnpCloseSocket(s);
     }
-    *s = INVALID_SOCKET;
+    s = INVALID_SOCKET;
 }
 
-static void closeSockets(MiniServerSockArray *out, int doclose)
+static void closeSockets(MiniServerSockArray *sockets, int doclose)
 {
 #ifdef INCLUDE_CLIENT_APIS
-    maybeCLoseAndInvalidate(&out->ssdpReqSock4, doclose);
-    maybeCLoseAndInvalidate(&out->ssdpReqSock6, doclose);
+        for (SOCKET& socket: sockets->ssdpReqSock4List) {
+            maybeCLoseAndInvalidate(socket, doclose);
+        }
+#ifdef UPNP_ENABLE_IPV6
+        if (using_ipv6()) {
+            for (SOCKET& socket: sockets->ssdpReqSock6List) {
+                maybeCLoseAndInvalidate(socket, doclose);
+            }
+        }
+#endif /* UPNP_ENABLE_IPV6 */
 #endif
-    maybeCLoseAndInvalidate(&out->ssdpSock4, doclose);
-    maybeCLoseAndInvalidate(&out->ssdpSock6, doclose);
-    maybeCLoseAndInvalidate(&out->ssdpSock6UlaGua, doclose);
+    maybeCLoseAndInvalidate(sockets->ssdpSock4, doclose);
+    maybeCLoseAndInvalidate(sockets->ssdpSock6, doclose);
+    maybeCLoseAndInvalidate(sockets->ssdpSock6UlaGua, doclose);
 }
 
-int get_ssdp_sockets(MiniServerSockArray *out, int port)
+int get_ssdp_req_socket_v4_for_netif(SOCKET& socket, const NetIF::Interface& netif, int port)
+{
+    std::string sadrv4;
+    const NetIF::IPAddr *ip = netif.firstipv4addr();
+
+    // No ip addr on iface ?
+    if (ip == nullptr) {
+        socket = INVALID_SOCKET;
+        return UPNP_E_SOCKET_ERROR;
+    }
+
+    sadrv4 = ip->straddr();
+    int retVal = create_ssdp_sock_reqv4(sadrv4, &socket, port);
+
+    // Failed to create socket ?
+    if (retVal != UPNP_E_SUCCESS) {
+        socket = INVALID_SOCKET;
+        return retVal;
+    }
+
+    return UPNP_E_SUCCESS;
+}
+
+#ifdef UPNP_ENABLE_IPV6
+int get_ssdp_req_socket_v6_for_netif(SOCKET& socket, const NetIF::Interface& netif, int port)
+{    
+    const NetIF::IPAddr* ip = netif.firstipv6addr();
+
+    // No ip addr on iface ?
+    if (ip == nullptr) {
+        socket = INVALID_SOCKET;
+        return UPNP_E_SOCKET_ERROR;
+    }
+
+    int retVal = create_ssdp_sock_reqv6(netif.getindex(), &socket, port);
+
+    // Failed to create socket ?
+    if (retVal != UPNP_E_SUCCESS) {
+        socket = INVALID_SOCKET;
+        return retVal;
+    }
+
+    return UPNP_E_SUCCESS;
+}
+#endif /* get_ssdp_req_socket_v6_for_netif */
+
+int get_ssdp_sockets(MiniServerSockArray *out, int startPort)
 {
     int retVal = UPNP_E_SOCKET_ERROR;
     bool hasIPV4 = !apiFirstIPV4Str().empty();
 
 #ifdef INCLUDE_CLIENT_APIS
+    // Create ipv4 and ipv6 CP search request sockets for each interface
+
+    out->ssdpReqSock4List.resize(g_netifs.size(), INVALID_SOCKET);
 #ifdef UPNP_ENABLE_IPV6
-    bool hasIPV6 = !apiFirstIPV6Str().empty();
-    if (using_ipv6()) {
-        /* Create the IPv6 socket for SSDP REQUESTS */
-        if (hasIPV6) {
-            if ((retVal = create_ssdp_sock_reqv6(&out->ssdpReqSock6, port)) != UPNP_E_SUCCESS) {
-                goto out;
-            }
-            /* For use by ssdp control point. */
-            gSsdpReqSocket6 = out->ssdpReqSock6;
-        }
-    }
+    out->ssdpReqSock6List.resize(g_netifs.size(), INVALID_SOCKET);
 #endif /* UPNP_ENABLE_IPV6 */
-    /* Create the IPv4 socket for SSDP REQUESTS */
-    if (hasIPV4) {
-        if ((retVal = create_ssdp_sock_reqv4(&out->ssdpReqSock4, port)) != UPNP_E_SUCCESS) {
-            goto out;
-        }
-        /* For use by ssdp control point. */
-        gSsdpReqSocket4 = out->ssdpReqSock4;
+
+    int netif_idx = 0;
+    int port = startPort;
+    for (const NetIF::Interface& netif: g_netifs) {
+        SOCKET& socket{out->ssdpReqSock4List[netif_idx]};
+        get_ssdp_req_socket_v4_for_netif(socket, netif, port);
+
+#ifdef UPNP_ENABLE_IPV6
+        if (using_ipv6()) {
+            SOCKET& sock6{out->ssdpReqSock6List[netif_idx]};
+            get_ssdp_req_socket_v6_for_netif(sock6, netif, port);
+        }            
+#endif /* UPNP_ENABLE_IPV6 */
+
+        port++;
+        netif_idx++;
     }
+
 #endif /* INCLUDE_CLIENT_APIS */
 
     /* Create the IPv4 socket for SSDP */
@@ -676,6 +723,7 @@ int get_ssdp_sockets(MiniServerSockArray *out, int port)
 #ifdef UPNP_ENABLE_IPV6
     if (using_ipv6()) {
         /* Create the IPv6 socket for SSDP */
+        bool hasIPV6 = !apiFirstIPV6Str().empty();
         if (hasIPV6) {
             if ((retVal = create_ssdp_sock_v6(false, &out->ssdpSock6)) != UPNP_E_SUCCESS) {
                 goto out;
