@@ -44,10 +44,11 @@
 #include "webserver.h"
 
 #include <algorithm>
+#include <cassert>
+#include <cinttypes>
 #include <map>
 #include <mutex>
-
-#include <cinttypes>
+#include <unordered_map>
 
 #include "genut.h"
 #include "ssdplib.h"
@@ -55,7 +56,6 @@
 #include "upnpapi.h"
 #include "uri.h"
 
-#include <cassert>
 #include <fcntl.h>
 #include <sys/stat.h>
 
@@ -112,7 +112,7 @@ struct SendInstruction {
  * module variables - Globals, static and externs.
  */
 
-static const std::map<std::string, const char*> gEncodedMediaTypes = {
+static const std::unordered_map<std::string_view, std::string_view> gEncodedMediaTypes = {
     {"aif", "audio/aiff"},
     {"aifc", "audio/aiff"},
     {"aiff", "audio/aiff"},
@@ -187,14 +187,17 @@ static const std::map<std::string, const char*> gEncodedMediaTypes = {
 
 
 /*! Global variable. A file system directory which serves as webserver
-    root. If this is not set from the API UpnpSetWebServerRootDir()
-    call, we do not serve files from the file system at all (only
-    possibly the virtual dir and/or the localDocs). */
+  root. If this is not set from the API UpnpSetWebServerRootDir()
+  call, we do not serve files from the file system at all (only
+  possibly the virtual dir and/or the localDocs). */
 static std::string gDocumentRootDir;
 
+/*! Global variable. A string which is set in the header field. */
+static std::string gWebServerCorsString;
+
 struct LocalDoc {
     std::string data;
-    time_t last_modified;
+    time_t last_modified{};
 };
 
 // Data which we serve directly: usually description
@@ -216,10 +219,9 @@ static std::mutex vdlmutex;
 
 
 /* Compute MIME type from file name extension. */
-static UPNP_INLINE int get_content_type(
-    const char *filename, std::string& content_type)
+static int get_content_type(const char* filename, std::string& content_type)
 {
-    const char *ctname = "application/octet-stream";
+    std::string_view ctname{"application/octet-stream"};
     content_type.clear();
     /* get ext */
     const char *e = strrchr(filename, '.');
@@ -235,15 +237,12 @@ static UPNP_INLINE int get_content_type(
     return 0;
 }
 
-int web_server_set_localdoc(
-    const std::string& path, const std::string& data, time_t last_modified)
+int web_server_set_localdoc(const std::string& path, const std::string& data, time_t last_modified)
 {
     if (path.empty() || path.front() != '/') {
         return UPNP_E_INVALID_PARAM;
     }
-    LocalDoc doc;
-    doc.data = data;
-    doc.last_modified = last_modified;
+    LocalDoc doc{data, last_modified};
     std::scoped_lock lck(gWebMutex);
     localDocs[path] = doc;
     return UPNP_E_SUCCESS;
@@ -280,10 +279,8 @@ static int get_file_info(const char *filename, struct File_Info *info)
     info->last_modified = s.st_mtime;
     int rc = get_content_type(filename, info->content_type);
     UpnpPrintf(UPNP_INFO, HTTP, __FILE__, __LINE__,
-               "get_file_info: %s, sz: %" PRIi64 ", mtime=%s rdable=%d\n",
-               filename, info->file_length,
-               make_date_string(info->last_modified).c_str(),
-               info->is_readable);
+               "get_file_info: %s, sz: %" PRIi64 ", mtime=%s rdable=%d\n", filename,
+               info->file_length, make_date_string(info->last_modified).c_str(), info->is_readable);
 
     return rc;
 }
@@ -299,6 +296,11 @@ int web_server_set_root_dir(const char *root_dir)
     return 0;
 }
 
+int web_server_set_cors(const char *cors_string)
+{
+    gWebServerCorsString = cors_string;
+    return 0;
+}
 
 int web_server_add_virtual_dir(
     const char *dirname, const void *cookie, const void **oldcookie)
@@ -307,8 +309,7 @@ int web_server_add_virtual_dir(
         return UPNP_E_INVALID_PARAM;
     }
 
-    UpnpPrintf(UPNP_DEBUG, HTTP, __FILE__, __LINE__,
-               "web_server_add_virtual_dir: [%s]\n", dirname);
+    UpnpPrintf(UPNP_DEBUG, HTTP, __FILE__, __LINE__, "web_server_add_virtual_dir: [%s]\n", dirname);
 
     VirtualDirListEntry entry;
     entry.cookie = cookie;
@@ -385,12 +386,11 @@ static const VirtualDirListEntry *isFileInVirtualDir(const std::string& path)
  * \brief Other header processing. Only HDR_ACCEPT_LANGUAGE for now.
  */
 static int CheckOtherHTTPHeaders(
-    MHDTransaction *mhdt, struct SendInstruction *RespInstr, int64_t)
+    const MHDTransaction* mhdt, struct SendInstruction* RespInstr, int64_t)
 {
-    for (const auto& header : mhdt->headers) {
+    for (const auto& [type, hvalue] : mhdt->headers) {
         /* find header type. */
-        int index = httpheader_str2int(header.first);
-        const std::string& hvalue = header.second;
+        int index = httpheader_str2int(type);
         if (index >= 0) {
             switch (index) {
             case HDR_ACCEPT_LANGUAGE:
@@ -493,10 +493,8 @@ static int process_request(
         std::string qs;
         if (!mhdt->queryvalues.empty()) {
             qs = "?";
-            for (const auto& entry : mhdt->queryvalues) {
-                qs += query_encode(entry.first) + "=" + query_encode(entry.second);
-                qs += "&";
-            }
+            for (const auto& [name, val] : mhdt->queryvalues)
+                qs += query_encode(name) + "=" + query_encode(val) + "&";
             qs.pop_back();
         }
         std::string bfilename{filename};
@@ -602,6 +600,9 @@ static int process_request(
     if (RespInstr->AcceptLanguageHeader[0] && WEB_SERVER_CONTENT_LANGUAGE[0]) {
         headers["content-language"] = WEB_SERVER_CONTENT_LANGUAGE;
     }
+    if (!gWebServerCorsString.empty()) {
+        headers["Access-Control-Allow-Origin"] = gWebServerCorsString;
+    }
     {
         std::string date = make_date_string(0);
         if (!date.empty())    
@@ -737,13 +738,12 @@ static void web_server_callback(MHDTransaction *mhdt)
     }
 
     bool serverhfound{false};
-    for (const auto& header : headers) {
-        //std::cerr << "web_server_callback: adding header [" << header.first <<
-        //"]->[" << header.second << "]\n";
-        if (!stringlowercmp("server", header.first)) {
+    for (const auto& [name, val] : headers) {
+        // std::cerr << "web_server_callback: adding header [" << name << "]->[" << val << "]\n";
+        if (!stringlowercmp("server", name)) {
             serverhfound = true;
         }
-        MHD_add_response_header(mhdt->response, header.first.c_str(), header.second.c_str());
+        MHD_add_response_header(mhdt->response, name.c_str(), val.c_str());
     }
     if (!serverhfound) {
         MHD_add_response_header(mhdt->response, "SERVER", get_sdk_device_info("").c_str());
